@@ -36,6 +36,7 @@ SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #include "i830.h"
 #include "i810_reg.h"
 #include "i915_drm.h"
+#include "brw_defines.h"
 #include <string.h>
 #include <sys/mman.h>
 #include <errno.h>
@@ -78,7 +79,31 @@ const int I830PatternROP[16] = {
 	ROP_1
 };
 
-static int uxa_pixmap_index;
+#if HAS_DEVPRIVATEKEYREC
+DevPrivateKeyRec uxa_pixmap_index;
+#else
+int uxa_pixmap_index;
+#endif
+
+static void
+ironlake_blt_workaround(ScrnInfoPtr scrn)
+{
+	intel_screen_private *intel = intel_get_screen_private(scrn);
+
+	/* Ironlake has a limitation that a 3D or Media command can't
+	 * be the first command after a BLT, unless it's
+	 * non-pipelined.  Instead of trying to track it and emit a
+	 * command at the right time, we just emit a dummy
+	 * non-pipelined 3D instruction after each blit.
+	 */
+
+	if (IS_IGDNG(intel)) {
+		BEGIN_BATCH(2);
+		OUT_BATCH(CMD_POLY_STIPPLE_OFFSET << 16);
+		OUT_BATCH(0);
+		ADVANCE_BATCH();
+	}
+}
 
 Bool
 i830_get_aperture_space(ScrnInfoPtr scrn, drm_intel_bo ** bo_table,
@@ -127,7 +152,7 @@ i830_uxa_pixmap_compute_size(PixmapPtr pixmap,
 {
 	ScrnInfoPtr scrn = xf86Screens[pixmap->drawable.pScreen->myNum];
 	intel_screen_private *intel = intel_get_screen_private(scrn);
-	int pitch_align;
+	int pitch, pitch_align;
 	int size;
 
 	if (*tiling != I915_TILING_NONE) {
@@ -135,10 +160,25 @@ i830_uxa_pixmap_compute_size(PixmapPtr pixmap,
 		pitch_align = intel->accel_pixmap_pitch_alignment;
 		size = ROUND_TO((w * pixmap->drawable.bitsPerPixel + 7) / 8,
 				pitch_align) * ALIGN (h, 2);
-		if (size < 4096)
+		if (!IS_I965G(intel)) {
+			/* Older hardware requires fences to be pot size
+			 * aligned with a minimum of 1 MiB, so causes
+			 * massive overallocation for small textures.
+			 */
+			if (size < 1024*1024/2)
+				*tiling = I915_TILING_NONE;
+		} else if (size <= 4096) {
+			/* Disable tiling beneath a page size, we will not see
+			 * any benefit from reducing TLB misses and instead
+			 * just incur extra cost when we require a fence.
+			 */
 			*tiling = I915_TILING_NONE;
+		}
 	}
 
+	pitch = (w * pixmap->drawable.bitsPerPixel + 7) / 8;
+	if (pitch <= 256)
+		*tiling = I915_TILING_NONE;
   repeat:
 	if (*tiling == I915_TILING_NONE) {
 		pitch_align = intel->accel_pixmap_pitch_alignment;
@@ -146,8 +186,7 @@ i830_uxa_pixmap_compute_size(PixmapPtr pixmap,
 		pitch_align = 512;
 	}
 
-	*stride = ROUND_TO((w * pixmap->drawable.bitsPerPixel + 7) / 8,
-			   pitch_align);
+	*stride = ROUND_TO(pitch, pitch_align);
 
 	if (*tiling == I915_TILING_NONE) {
 		/* Round the height up so that the GPU's access to a 2x2 aligned
@@ -156,7 +195,7 @@ i830_uxa_pixmap_compute_size(PixmapPtr pixmap,
 		 */
 		size = *stride * ALIGN(h, 2);
 	} else {
-		int aligned_h = h;
+		int aligned_h;
 		if (*tiling == I915_TILING_X)
 			aligned_h = ALIGN(h, 8);
 		else
@@ -181,6 +220,35 @@ i830_uxa_pixmap_compute_size(PixmapPtr pixmap,
 	return size;
 }
 
+static Bool
+i830_uxa_check_solid(DrawablePtr drawable, int alu, Pixel planemask)
+{
+	ScrnInfoPtr scrn = xf86Screens[drawable->pScreen->myNum];
+	intel_screen_private *intel = intel_get_screen_private(scrn);
+
+	if (IS_GEN6(intel)) {
+		intel_debug_fallback(scrn,
+				     "Sandybridge BLT engine not supported\n");
+		return FALSE;
+	}
+
+	if (!UXA_PM_IS_SOLID(drawable, planemask)) {
+		intel_debug_fallback(scrn, "planemask is not solid\n");
+		return FALSE;
+	}
+
+	switch (drawable->bitsPerPixel) {
+	case 8:
+	case 16:
+	case 32:
+		break;
+	default:
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
 /**
  * Sets up hardware state for a series of solid fills.
  */
@@ -189,32 +257,14 @@ i830_uxa_prepare_solid(PixmapPtr pixmap, int alu, Pixel planemask, Pixel fg)
 {
 	ScrnInfoPtr scrn = xf86Screens[pixmap->drawable.pScreen->myNum];
 	intel_screen_private *intel = intel_get_screen_private(scrn);
-	unsigned long pitch;
 	drm_intel_bo *bo_table[] = {
 		NULL,		/* batch_bo */
 		i830_get_pixmap_bo(pixmap),
 	};
 
-	if (!UXA_PM_IS_SOLID(&pixmap->drawable, planemask)) {
-		intel_debug_fallback(scrn, "planemask is not solid\n");
-		return FALSE;
-	}
-
-	if (pixmap->drawable.bitsPerPixel == 24) {
-		intel_debug_fallback(scrn, "solid 24bpp unsupported!\n");
-		return FALSE;
-	}
-
-	if (pixmap->drawable.bitsPerPixel < 8) {
-		intel_debug_fallback(scrn, "under 8bpp pixmaps unsupported\n");
-		return FALSE;
-	}
-
 	if (!intel_check_pitch_2d(pixmap))
 		return FALSE;
 
-	pitch = i830_pixmap_pitch(pixmap);
-
 	if (!i830_pixmap_pitch_is_aligned(pixmap)) {
 		intel_debug_fallback(scrn, "pixmap pitch not aligned");
 		return FALSE;
@@ -237,6 +287,7 @@ i830_uxa_prepare_solid(PixmapPtr pixmap, int alu, Pixel planemask, Pixel fg)
 		break;
 	}
 	intel->BR[16] = fg;
+
 	return TRUE;
 }
 
@@ -247,6 +298,18 @@ static void i830_uxa_solid(PixmapPtr pixmap, int x1, int y1, int x2, int y2)
 	unsigned long pitch;
 	uint32_t cmd;
 
+	if (x1 < 0)
+		x1 = 0;
+	if (y1 < 0)
+		y1 = 0;
+	if (x2 > pixmap->drawable.width)
+		x2 = pixmap->drawable.width;
+	if (y2 > pixmap->drawable.height)
+		y2 = pixmap->drawable.height;
+
+	if (x2 <= x1 || y2 <= y1)
+		return;
+
 	pitch = i830_pixmap_pitch(pixmap);
 
 	{
@@ -269,11 +332,13 @@ static void i830_uxa_solid(PixmapPtr pixmap, int x1, int y1, int x2, int y2)
 		OUT_BATCH(intel->BR[13] | pitch);
 		OUT_BATCH((y1 << 16) | (x1 & 0xffff));
 		OUT_BATCH((y2 << 16) | (x2 & 0xffff));
-		OUT_RELOC_PIXMAP(pixmap, I915_GEM_DOMAIN_RENDER,
-				 I915_GEM_DOMAIN_RENDER, 0);
+		OUT_RELOC_PIXMAP_FENCED(pixmap, I915_GEM_DOMAIN_RENDER,
+					I915_GEM_DOMAIN_RENDER, 0);
 		OUT_BATCH(intel->BR[16]);
 		ADVANCE_BATCH();
 	}
+
+	ironlake_blt_workaround(scrn);
 }
 
 static void i830_uxa_done_solid(PixmapPtr pixmap)
@@ -287,6 +352,45 @@ static void i830_uxa_done_solid(PixmapPtr pixmap)
  * TODO:
  *   - support planemask using FULL_BLT_CMD?
  */
+static Bool
+i830_uxa_check_copy(PixmapPtr source, PixmapPtr dest,
+		    int alu, Pixel planemask)
+{
+	ScrnInfoPtr scrn = xf86Screens[dest->drawable.pScreen->myNum];
+	intel_screen_private *intel = intel_get_screen_private(scrn);
+
+	if (IS_GEN6(intel)) {
+		intel_debug_fallback(scrn,
+				     "Sandybridge BLT engine not supported\n");
+		return FALSE;
+	}
+
+	if (!UXA_PM_IS_SOLID(&source->drawable, planemask)) {
+		intel_debug_fallback(scrn, "planemask is not solid");
+		return FALSE;
+	}
+
+	if (source->drawable.bitsPerPixel != dest->drawable.bitsPerPixel) {
+		intel_debug_fallback(scrn, "mixed bpp copies unsupported\n");
+		return FALSE;
+	}
+	switch (source->drawable.bitsPerPixel) {
+	case 8:
+	case 16:
+	case 32:
+		break;
+	default:
+		return FALSE;
+	}
+
+	if (!intel_check_pitch_2d(source))
+		return FALSE;
+	if (!intel_check_pitch_2d(dest))
+		return FALSE;
+
+	return TRUE;
+}
+
 static Bool
 i830_uxa_prepare_copy(PixmapPtr source, PixmapPtr dest, int xdir,
 		      int ydir, int alu, Pixel planemask)
@@ -299,28 +403,12 @@ i830_uxa_prepare_copy(PixmapPtr source, PixmapPtr dest, int xdir,
 		i830_get_pixmap_bo(dest),
 	};
 
-	if (!UXA_PM_IS_SOLID(&source->drawable, planemask)) {
-		intel_debug_fallback(scrn, "planemask is not solid");
-		return FALSE;
-	}
-
-	if (dest->drawable.bitsPerPixel < 8) {
-		intel_debug_fallback(scrn, "under 8bpp pixmaps unsupported\n");
-		return FALSE;
-	}
-
 	if (!i830_get_aperture_space(scrn, bo_table, ARRAY_SIZE(bo_table)))
 		return FALSE;
 
-	if (!intel_check_pitch_2d(source))
-		return FALSE;
-	if (!intel_check_pitch_2d(dest))
-		return FALSE;
-
 	intel->render_source = source;
 
 	intel->BR[13] = I830CopyROP[alu] << 16;
-
 	switch (source->drawable.bitsPerPixel) {
 	case 8:
 		break;
@@ -331,6 +419,7 @@ i830_uxa_prepare_copy(PixmapPtr source, PixmapPtr dest, int xdir,
 		intel->BR[13] |= ((1 << 25) | (1 << 24));
 		break;
 	}
+
 	return TRUE;
 }
 
@@ -379,18 +468,19 @@ i830_uxa_copy(PixmapPtr dest, int src_x1, int src_y1, int dst_x1,
 		OUT_BATCH(intel->BR[13] | dst_pitch);
 		OUT_BATCH((dst_y1 << 16) | (dst_x1 & 0xffff));
 		OUT_BATCH((dst_y2 << 16) | (dst_x2 & 0xffff));
-		OUT_RELOC_PIXMAP(dest,
-				 I915_GEM_DOMAIN_RENDER,
-				 I915_GEM_DOMAIN_RENDER,
-				 0);
+		OUT_RELOC_PIXMAP_FENCED(dest,
+					I915_GEM_DOMAIN_RENDER,
+					I915_GEM_DOMAIN_RENDER,
+					0);
 		OUT_BATCH((src_y1 << 16) | (src_x1 & 0xffff));
 		OUT_BATCH(src_pitch);
-		OUT_RELOC_PIXMAP(intel->render_source,
-				 I915_GEM_DOMAIN_RENDER, 0,
-				 0);
+		OUT_RELOC_PIXMAP_FENCED(intel->render_source,
+					I915_GEM_DOMAIN_RENDER, 0,
+					0);
 
 		ADVANCE_BATCH();
 	}
+	ironlake_blt_workaround(scrn);
 }
 
 static void i830_uxa_done_copy(PixmapPtr dest)
@@ -408,6 +498,10 @@ static void i830_uxa_done_copy(PixmapPtr dest)
 void i830_done_composite(PixmapPtr dest)
 {
 	ScrnInfoPtr scrn = xf86Screens[dest->drawable.pScreen->myNum];
+	intel_screen_private *intel = intel_get_screen_private(scrn);
+
+	if (intel->vertex_flush)
+		intel->vertex_flush(intel);
 
 	i830_debug_flush(scrn);
 }
@@ -493,16 +587,6 @@ Bool i830_transform_is_affine(PictTransformPtr t)
 	return t->matrix[2][0] == 0 && t->matrix[2][1] == 0;
 }
 
-struct intel_pixmap *i830_get_pixmap_intel(PixmapPtr pixmap)
-{
-	return dixLookupPrivate(&pixmap->devPrivates, &uxa_pixmap_index);
-}
-
-static void i830_uxa_set_pixmap_intel(PixmapPtr pixmap, struct intel_pixmap *intel)
-{
-	dixSetPrivate(&pixmap->devPrivates, &uxa_pixmap_index, intel);
-}
-
 dri_bo *i830_get_pixmap_bo(PixmapPtr pixmap)
 {
 	struct intel_pixmap *intel;
@@ -516,17 +600,32 @@ dri_bo *i830_get_pixmap_bo(PixmapPtr pixmap)
 
 void i830_set_pixmap_bo(PixmapPtr pixmap, dri_bo * bo)
 {
+	ScrnInfoPtr scrn = xf86Screens[pixmap->drawable.pScreen->myNum];
+	intel_screen_private *intel = intel_get_screen_private(scrn);
 	struct intel_pixmap *priv;
 
 	priv = i830_get_pixmap_intel(pixmap);
 
-	if (priv != NULL) {
-		dri_bo_unreference(priv->bo);
+	if (priv == NULL && bo == NULL)
+	    return;
 
-		priv->flush_read_domains = priv->flush_write_domain = 0;
-		priv->batch_read_domains = priv->batch_write_domain = 0;
-		list_del(&priv->batch);
-		list_del(&priv->flush);
+	if (priv != NULL) {
+		if (priv->bo == bo)
+			return;
+
+		if (list_is_empty(&priv->batch)) {
+			dri_bo_unreference(priv->bo);
+		} else if (!drm_intel_bo_is_reusable(priv->bo)) {
+			dri_bo_unreference(priv->bo);
+			list_del(&priv->batch);
+			list_del(&priv->flush);
+		} else {
+			list_add(&priv->in_flight, &intel->in_flight);
+			priv = NULL;
+		}
+
+		if (intel->render_current_dest == pixmap)
+		    intel->render_current_dest = NULL;
 	}
 
 	if (bo != NULL) {
@@ -534,7 +633,7 @@ void i830_set_pixmap_bo(PixmapPtr pixmap, dri_bo * bo)
 		int ret;
 
 		if (priv == NULL) {
-			priv = xcalloc(1, sizeof (struct intel_pixmap));
+			priv = calloc(1, sizeof (struct intel_pixmap));
 			if (priv == NULL)
 				goto BAIL;
 
@@ -544,6 +643,7 @@ void i830_set_pixmap_bo(PixmapPtr pixmap, dri_bo * bo)
 
 		dri_bo_reference(bo);
 		priv->bo = bo;
+		priv->stride = i830_pixmap_pitch(pixmap);
 
 		ret = drm_intel_bo_get_tiling(bo,
 					      &priv->tiling,
@@ -554,13 +654,13 @@ void i830_set_pixmap_bo(PixmapPtr pixmap, dri_bo * bo)
 		}
 	} else {
 		if (priv != NULL) {
-			xfree(priv);
+			free(priv);
 			priv = NULL;
 		}
 	}
 
   BAIL:
-	i830_uxa_set_pixmap_intel(pixmap, priv);
+	i830_set_pixmap_intel(pixmap, priv);
 }
 
 static Bool i830_uxa_prepare_access(PixmapPtr pixmap, uxa_access_t access)
@@ -575,8 +675,7 @@ static Bool i830_uxa_prepare_access(PixmapPtr pixmap, uxa_access_t access)
 	    (access == UXA_ACCESS_RW || priv->batch_write_domain))
 		intel_batch_submit(scrn);
 
-	/* No VT sema or GEM?  No GTT mapping. */
-	if (!scrn->vtSema || bo->size > intel->max_gtt_map_size) {
+	if (bo->size > intel->max_gtt_map_size) {
 		ret = dri_bo_map(bo, access == UXA_ACCESS_RW);
 		if (ret != 0) {
 			xf86DrvMsg(scrn->scrnIndex, X_WARNING,
@@ -604,22 +703,39 @@ static Bool i830_uxa_prepare_access(PixmapPtr pixmap, uxa_access_t access)
 static void i830_uxa_finish_access(PixmapPtr pixmap)
 {
 	dri_bo *bo = i830_get_pixmap_bo(pixmap);
+	ScreenPtr screen = pixmap->drawable.pScreen;
+	ScrnInfoPtr scrn = xf86Screens[screen->myNum];
+	intel_screen_private *intel = intel_get_screen_private(scrn);
 
-	if (bo) {
-		ScreenPtr screen = pixmap->drawable.pScreen;
-		ScrnInfoPtr scrn = xf86Screens[screen->myNum];
-		intel_screen_private *intel = intel_get_screen_private(scrn);
+	if (bo == intel->front_buffer)
+		intel->need_flush = TRUE;
 
-		if (bo == intel->front_buffer->bo)
-			intel->need_flush = TRUE;
+	if (bo->size > intel->max_gtt_map_size)
+		dri_bo_unmap(bo);
+	else
+		drm_intel_gem_bo_unmap_gtt(bo);
 
-		if (!scrn->vtSema || bo->size > intel->max_gtt_map_size)
-			dri_bo_unmap(bo);
-		else
-			drm_intel_gem_bo_unmap_gtt(bo);
+	pixmap->devPrivate.ptr = NULL;
+}
 
-		pixmap->devPrivate.ptr = NULL;
+static Bool i830_bo_put_image(PixmapPtr pixmap, dri_bo *bo, char *src, int src_pitch, int w, int h)
+{
+	int stride = i830_pixmap_pitch(pixmap);
+
+	if (src_pitch == stride) {
+		memcpy (bo->virtual, src, stride * h);
+	} else {
+		char *dst = bo->virtual;
+		int row_length = w * pixmap->drawable.bitsPerPixel/8;
+		int num_rows = h;
+		while (num_rows--) {
+			memcpy (dst, src, row_length);
+			src += src_pitch;
+			dst += stride;
+		}
 	}
+
+	return TRUE;
 }
 
 static Bool
@@ -634,6 +750,7 @@ i830_uxa_pixmap_swap_bo_with_image(PixmapPtr pixmap,
 	int stride;
 	int w = pixmap->drawable.width;
 	int h = pixmap->drawable.height;
+	Bool ret;
 
 	priv = i830_get_pixmap_intel(pixmap);
 
@@ -652,17 +769,12 @@ i830_uxa_pixmap_swap_bo_with_image(PixmapPtr pixmap,
 		if (tiling != I915_TILING_NONE)
 			drm_intel_bo_set_tiling(bo, &tiling, stride);
 
-		dri_bo_unreference(priv->bo);
-		priv->bo = bo;
-		priv->tiling = tiling;
-		priv->batch_read_domains = priv->batch_write_domain = 0;
-		priv->flush_read_domains = priv->flush_write_domain = 0;
-		list_del(&priv->batch);
-		list_del(&priv->flush);
 		pixmap->drawable.pScreen->ModifyPixmapHeader(pixmap,
 							     w, h,
 							     0, 0,
 							     stride, NULL);
+		i830_set_pixmap_bo(pixmap, bo);
+		dri_bo_unreference(bo);
 	} else {
 		bo = priv->bo;
 		stride = i830_pixmap_pitch(pixmap);
@@ -674,22 +786,11 @@ i830_uxa_pixmap_swap_bo_with_image(PixmapPtr pixmap,
 		return FALSE;
 	}
 
-	if (src_pitch == stride) {
-		memcpy (bo->virtual, src, src_pitch * h);
-	} else {
-		char *dst = bo->virtual;
-
-		w *= pixmap->drawable.bitsPerPixel/8;
-		while (h--) {
-			memcpy (dst, src, w);
-			src += src_pitch;
-			dst += stride;
-		}
-	}
+	ret = i830_bo_put_image(pixmap, bo, src, src_pitch, w, h);
 
 	drm_intel_gem_bo_unmap_gtt(bo);
 
-	return TRUE;
+	return ret;
 }
 
 static Bool i830_uxa_put_image(PixmapPtr pixmap,
@@ -701,7 +802,6 @@ static Bool i830_uxa_put_image(PixmapPtr pixmap,
 	ScrnInfoPtr scrn = xf86Screens[screen->myNum];
 	PixmapPtr scratch;
 	struct intel_pixmap *priv;
-	Bool scratch_pixmap;
 	GCPtr gc;
 	Bool ret;
 
@@ -715,9 +815,9 @@ static Bool i830_uxa_put_image(PixmapPtr pixmap,
 	}
 
 	priv = i830_get_pixmap_intel(pixmap);
-	if (priv->batch_read_domains || drm_intel_bo_busy(priv->bo)) {
+	if (priv->batch_read_domains ||
+	    drm_intel_bo_busy(priv->bo)) {
 		dri_bo *bo;
-		int stride;
 
 		/* Partial replacement, copy incoming image to a bo and blit. */
 		scratch = (*screen->CreatePixmap)(screen, w, h,
@@ -734,49 +834,77 @@ static Bool i830_uxa_put_image(PixmapPtr pixmap,
 			return FALSE;
 		}
 
-		stride = i830_pixmap_pitch(scratch);
-		if (src_pitch == stride) {
-			memcpy (bo->virtual, src, stride * h);
-		} else {
-			char *dst = bo->virtual;
-			int row_length = w * pixmap->drawable.bitsPerPixel/8;
-			int num_rows = h;
-			while (num_rows--) {
-				memcpy (dst, src, row_length);
-				src += src_pitch;
-				dst += stride;
+		ret = i830_bo_put_image(scratch, bo, src, src_pitch, w, h);
+
+		drm_intel_gem_bo_unmap_gtt(bo);
+
+		if (ret) {
+			gc = GetScratchGC(pixmap->drawable.depth, screen);
+			if (gc) {
+				ValidateGC(&pixmap->drawable, gc);
+
+				(*gc->ops->CopyArea)(&scratch->drawable,
+						     &pixmap->drawable,
+						     gc, 0, 0, w, h, x, y);
+
+				FreeScratchGC(gc);
+			} else
+				ret = FALSE;
+		}
+
+		(*screen->DestroyPixmap)(scratch);
+	} else {
+		int dst_pitch;
+
+		/* bo is not busy so can be mapped without a stall, upload in-place. */
+		if (drm_intel_gem_bo_map_gtt(priv->bo)) {
+			xf86DrvMsg(scrn->scrnIndex, X_WARNING,
+				   "%s: bo map failed\n", __FUNCTION__);
+			return FALSE;
+		}
+
+		ret = TRUE;
+
+		/* Older version of pixman did not allow blt for <= 8bpp
+		 * images, so if the blt fails fallback to using the fb.
+		 */
+		dst_pitch = i830_pixmap_pitch (pixmap);
+		if (! pixman_blt((uint32_t *)src,
+				 priv->bo->virtual,
+				 src_pitch / sizeof(uint32_t),
+				 dst_pitch / sizeof(uint32_t),
+				 pixmap->drawable.bitsPerPixel,
+				 pixmap->drawable.bitsPerPixel,
+				 0, 0,
+				 x, y,
+				 w, h))
+		{
+			ret = FALSE;
+
+			scratch = GetScratchPixmapHeader(screen,
+							 w, h,
+							 pixmap->drawable.depth,
+							 pixmap->drawable.bitsPerPixel,
+							 src_pitch,
+							 (pointer) src);
+			if (scratch) {
+				gc = GetScratchGC(pixmap->drawable.depth, screen);
+				if (gc) {
+					ValidateGC(&pixmap->drawable, gc);
+					pixmap->devPrivate.ptr = priv->bo->virtual;
+					ret = !! fbCopyArea(&scratch->drawable, &pixmap->drawable, gc,
+							    0, 0,
+							    w, h,
+							    x, y);
+					FreeScratchGC(gc);
+				}
+				FreeScratchPixmapHeader(scratch);
 			}
 		}
 
-		drm_intel_gem_bo_unmap_gtt(bo);
-		scratch_pixmap = FALSE;
-	} else {
-		/* bo is not busy so can be mapped without a stall, upload in-place. */
-		scratch = GetScratchPixmapHeader(screen, w, h,
-						 pixmap->drawable.depth,
-						 pixmap->drawable.bitsPerPixel,
-						 src_pitch, src);
-		scratch_pixmap = TRUE;
+		drm_intel_gem_bo_unmap_gtt(priv->bo);
 	}
 
-	ret = FALSE;
-	gc = GetScratchGC(pixmap->drawable.depth, screen);
-	if (gc) {
-		ValidateGC(&pixmap->drawable, gc);
-
-		(*gc->ops->CopyArea)(&scratch->drawable,
-				     &pixmap->drawable,
-				     gc, 0, 0, w, h, x, y);
-
-		FreeScratchGC(gc);
-		ret = TRUE;
-	}
-
-	if (scratch_pixmap)
-		FreeScratchPixmapHeader(scratch);
-	else
-		(*screen->DestroyPixmap)(scratch);
-
 	return ret;
 }
 
@@ -786,7 +914,7 @@ void i830_uxa_block_handler(ScreenPtr screen)
 	intel_screen_private *intel = intel_get_screen_private(scrn);
 
 	if (intel->need_flush) {
-		dri_bo_wait_rendering(intel->front_buffer->bo);
+		dri_bo_wait_rendering(intel->front_buffer);
 		intel->need_flush = FALSE;
 	}
 }
@@ -813,43 +941,85 @@ i830_uxa_create_pixmap(ScreenPtr screen, int w, int h, int depth,
 	if (w > 32767 || h > 32767)
 		return NullPixmap;
 
-	if (usage == CREATE_PIXMAP_USAGE_GLYPH_PICTURE)
+	if (depth == 1 || intel->force_fallback)
+		return fbCreatePixmap(screen, w, h, depth, usage);
+
+	if (usage == CREATE_PIXMAP_USAGE_GLYPH_PICTURE && w <= 32 && h <= 32)
 		return fbCreatePixmap(screen, w, h, depth, usage);
 
 	pixmap = fbCreatePixmap(screen, 0, 0, depth, usage);
 
 	if (w && h) {
 		struct intel_pixmap *priv;
-		unsigned int size;
+		unsigned int size, tiling;
 		int stride;
 
-		priv = xcalloc(1, sizeof (struct intel_pixmap));
-		if (priv == NULL) {
+		/* Always attempt to tile, compute_size() will remove the
+		 * tiling for pixmaps that are either too large or too small
+		 * to be effectively tiled.
+		 */
+		tiling = I915_TILING_X;
+		if (usage == INTEL_CREATE_PIXMAP_TILING_Y)
+			tiling = I915_TILING_Y;
+		if (usage == UXA_CREATE_PIXMAP_FOR_MAP)
+			tiling = I915_TILING_NONE;
+
+		if (tiling != I915_TILING_NONE) {
+		    if (h <= 4)
+			tiling = I915_TILING_NONE;
+		    if (h <= 16 && tiling == I915_TILING_Y)
+			tiling = I915_TILING_X;
+		}
+		size = i830_uxa_pixmap_compute_size(pixmap, w, h, &tiling, &stride);
+
+		/* Fail very large allocations.  Large BOs will tend to hit SW fallbacks
+		 * frequently, and also will tend to fail to successfully map when doing
+		 * SW fallbacks because we overcommit address space for BO access.
+		 */
+		if (size > intel->max_bo_size) {
 			fbDestroyPixmap(pixmap);
-			return NullPixmap;
+			return fbCreatePixmap(screen, w, h, depth, usage);
 		}
 
-		if (usage == INTEL_CREATE_PIXMAP_TILING_X)
-			priv->tiling = I915_TILING_X;
-		else if (usage == INTEL_CREATE_PIXMAP_TILING_Y)
-			priv->tiling = I915_TILING_Y;
-		else
-			priv->tiling = I915_TILING_NONE;
+		/* Perform a preliminary search for an in-flight bo */
+		if (usage != UXA_CREATE_PIXMAP_FOR_MAP) {
+			int aligned_h;
 
-		size = i830_uxa_pixmap_compute_size(pixmap, w, h,
-						    &priv->tiling, &stride);
+			if (tiling == I915_TILING_X)
+				aligned_h = ALIGN(h, 8);
+			else if (tiling == I915_TILING_Y)
+				aligned_h = ALIGN(h, 32);
+			else
+				aligned_h = ALIGN(h, 2);
 
-		/* Fail very large allocations on 32-bit systems.  Large BOs will
-		 * tend to hit SW fallbacks frequently, and also will tend to fail
-		 * to successfully map when doing SW fallbacks because we overcommit
-		 * address space for BO access.
-		 *
-		 * Note that size should fit in 32 bits.  We throw out >32767x32767x4,
-		 * and pitch alignment could get us up to 32768x32767x4.
-		 */
-		if (sizeof(unsigned long) == 4 &&
-		    size > (unsigned int)(1024 * 1024 * 1024)) {
-			xfree(priv);
+			list_foreach_entry(priv, struct intel_pixmap,
+					   &intel->in_flight,
+					   in_flight) {
+				if (priv->tiling != tiling)
+					continue;
+
+				if (tiling == I915_TILING_NONE) {
+				    if (priv->bo->size < size)
+					    continue;
+
+					priv->stride = stride;
+				} else {
+					if (priv->stride < stride ||
+					    priv->bo->size < priv->stride * aligned_h)
+						continue;
+
+					stride = priv->stride;
+				}
+
+				list_del(&priv->in_flight);
+				screen->ModifyPixmapHeader(pixmap, w, h, 0, 0, stride, NULL);
+				i830_set_pixmap_intel(pixmap, priv);
+				return pixmap;
+			}
+		}
+
+		priv = calloc(1, sizeof (struct intel_pixmap));
+		if (priv == NULL) {
 			fbDestroyPixmap(pixmap);
 			return NullPixmap;
 		}
@@ -862,11 +1032,13 @@ i830_uxa_create_pixmap(ScreenPtr screen, int w, int h, int depth,
 								 "pixmap",
 								 size, 0);
 		if (!priv->bo) {
-			xfree(priv);
+			free(priv);
 			fbDestroyPixmap(pixmap);
 			return NullPixmap;
 		}
 
+		priv->stride = stride;
+		priv->tiling = tiling;
 		if (priv->tiling != I915_TILING_NONE)
 			drm_intel_bo_set_tiling(priv->bo,
 						&priv->tiling,
@@ -876,7 +1048,7 @@ i830_uxa_create_pixmap(ScreenPtr screen, int w, int h, int depth,
 
 		list_init(&priv->batch);
 		list_init(&priv->flush);
-		i830_uxa_set_pixmap_intel(pixmap, priv);
+		i830_set_pixmap_intel(pixmap, priv);
 	}
 
 	return pixmap;
@@ -895,7 +1067,7 @@ void i830_uxa_create_screen_resources(ScreenPtr screen)
 {
 	ScrnInfoPtr scrn = xf86Screens[screen->myNum];
 	intel_screen_private *intel = intel_get_screen_private(scrn);
-	dri_bo *bo = intel->front_buffer->bo;
+	dri_bo *bo = intel->front_buffer;
 
 	if (bo != NULL) {
 		PixmapPtr pixmap = screen->GetScreenPixmap(screen);
@@ -908,7 +1080,11 @@ Bool i830_uxa_init(ScreenPtr screen)
 	ScrnInfoPtr scrn = xf86Screens[screen->myNum];
 	intel_screen_private *intel = intel_get_screen_private(scrn);
 
+#if HAS_DIXREGISTERPRIVATEKEY
+	if (!dixRegisterPrivateKey(&uxa_pixmap_index, PRIVATE_PIXMAP, 0))
+#else
 	if (!dixRequestPrivate(&uxa_pixmap_index, 0))
+#endif
 		return FALSE;
 
 	intel->uxa_driver = uxa_driver_alloc();
@@ -917,16 +1093,27 @@ Bool i830_uxa_init(ScreenPtr screen)
 
 	memset(intel->uxa_driver, 0, sizeof(*intel->uxa_driver));
 
+	intel->force_fallback = FALSE;
+
 	intel->bufferOffset = 0;
 	intel->uxa_driver->uxa_major = 1;
 	intel->uxa_driver->uxa_minor = 0;
 
+	intel->render_current_dest = NULL;
+	intel->prim_offset = 0;
+	intel->vertex_count = 0;
+	intel->floats_per_vertex = 0;
+	intel->last_floats_per_vertex = 0;
+	intel->vertex_bo = NULL;
+
 	/* Solid fill */
+	intel->uxa_driver->check_solid = i830_uxa_check_solid;
 	intel->uxa_driver->prepare_solid = i830_uxa_prepare_solid;
 	intel->uxa_driver->solid = i830_uxa_solid;
 	intel->uxa_driver->done_solid = i830_uxa_done_solid;
 
 	/* Copy */
+	intel->uxa_driver->check_copy = i830_uxa_check_copy;
 	intel->uxa_driver->prepare_copy = i830_uxa_prepare_copy;
 	intel->uxa_driver->copy = i830_uxa_copy;
 	intel->uxa_driver->done_copy = i830_uxa_done_copy;
@@ -934,17 +1121,22 @@ Bool i830_uxa_init(ScreenPtr screen)
 	/* Composite */
 	if (!IS_I9XX(intel)) {
 		intel->uxa_driver->check_composite = i830_check_composite;
+		intel->uxa_driver->check_composite_target = i830_check_composite_target;
+		intel->uxa_driver->check_composite_texture = i830_check_composite_texture;
 		intel->uxa_driver->prepare_composite = i830_prepare_composite;
 		intel->uxa_driver->composite = i830_composite;
 		intel->uxa_driver->done_composite = i830_done_composite;
 	} else if (IS_I915G(intel) || IS_I915GM(intel) ||
 		   IS_I945G(intel) || IS_I945GM(intel) || IS_G33CLASS(intel)) {
 		intel->uxa_driver->check_composite = i915_check_composite;
+		intel->uxa_driver->check_composite_target = i915_check_composite_target;
+		intel->uxa_driver->check_composite_texture = i915_check_composite_texture;
 		intel->uxa_driver->prepare_composite = i915_prepare_composite;
 		intel->uxa_driver->composite = i915_composite;
 		intel->uxa_driver->done_composite = i830_done_composite;
 	} else {
 		intel->uxa_driver->check_composite = i965_check_composite;
+		intel->uxa_driver->check_composite_texture = i965_check_composite_texture;
 		intel->uxa_driver->prepare_composite = i965_prepare_composite;
 		intel->uxa_driver->composite = i965_composite;
 		intel->uxa_driver->done_composite = i830_done_composite;
@@ -957,16 +1149,16 @@ Bool i830_uxa_init(ScreenPtr screen)
 	intel->uxa_driver->finish_access = i830_uxa_finish_access;
 	intel->uxa_driver->pixmap_is_offscreen = i830_uxa_pixmap_is_offscreen;
 
+	screen->CreatePixmap = i830_uxa_create_pixmap;
+	screen->DestroyPixmap = i830_uxa_destroy_pixmap;
+
 	if (!uxa_driver_init(screen, intel->uxa_driver)) {
 		xf86DrvMsg(scrn->scrnIndex, X_ERROR,
 			   "UXA initialization failed\n");
-		xfree(intel->uxa_driver);
+		free(intel->uxa_driver);
 		return FALSE;
 	}
 
-	screen->CreatePixmap = i830_uxa_create_pixmap;
-	screen->DestroyPixmap = i830_uxa_destroy_pixmap;
-
 	uxa_set_fallback_debug(screen, intel->fallback_debug);
 
 	return TRUE;
